from typing import Tuple, Type

from pydantic import BaseModel

from pyappconf.model import BaseConfig
from tests.config import JSON_PATH, SCHEMA_JSON_PATH, TOML_PATH, YAML_PATH
from tests.fixtures.model import model_classes, model_object


def test_to_json(model_object: BaseConfig):
    assert model_object.to_json() == JSON_PATH.read_text()


def test_to_yaml(model_object: BaseConfig):
    assert model_object.to_yaml() == YAML_PATH.read_text()


def test_to_toml(model_object: BaseConfig):
    assert model_object.to_toml() == TOML_PATH.read_text()


def test_from_json(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, SubModel = model_classes
    loaded_object = MyConfig.parse_json(JSON_PATH)
    assert model_object == loaded_object


def test_from_yaml(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, SubModel = model_classes
    loaded_object = MyConfig.parse_yaml(YAML_PATH)
    assert model_object == loaded_object


def test_from_toml(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, SubModel = model_classes
    loaded_object = MyConfig.parse_toml(TOML_PATH)
    assert model_object == loaded_object


def test_json_schema(model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]):
    MyConfig, SubModel = model_classes
    assert MyConfig.schema_json(indent=2) == SCHEMA_JSON_PATH.read_text()



import datetime
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Tuple
from uuid import UUID

import pytest
from pydantic import BaseModel, Field

from tests.fixtures.data import get_default_data


class MyEnum(str, Enum):
    ONE = "one"
    TWO = "two"


class SubModel(BaseModel):
    a: str
    b: float


class MyModel(BaseModel):
    string: str
    integer: int
    custom: SubModel
    dictionary: Dict[str, SubModel]
    str_list: List[str]
    int_tuple: Tuple[int, ...]
    uuid: UUID
    date: datetime.date
    time: datetime.datetime

    default_string: str = "woo"
    default_custom: SubModel = SubModel(a="yeah", b=5.6)
    default_enum: MyEnum = MyEnum.ONE
    default_enum_list: List[MyEnum] = Field(
        default_factory=lambda: [MyEnum.ONE, MyEnum.TWO]
    )
    default_uuid_list: List[UUID] = Field(
        default_factory=lambda: [UUID("a" * 32), UUID("b" * 32)]
    )
    default_time_with_tz: datetime.datetime = datetime.datetime(
        2022, 1, 1, tzinfo=datetime.timezone.utc
    )
    default_time_list: List[datetime.datetime] = Field(
        default_factory=lambda: [
            datetime.datetime(2022, 1, 1, tzinfo=datetime.timezone.min),
            datetime.datetime(2022, 1, 2, tzinfo=datetime.timezone.max),
        ]
    )
    default_date_list: List[datetime.date] = Field(
        default_factory=lambda: [
            datetime.date(2022, 1, 1),
            datetime.date(2022, 1, 2),
        ]
    )
    file_path: Path = Path("/a/b.txt")


def default_func():
    return "default"


class RequiredCallableClass:
    def __call__(self):
        return "required"

    def __pyconfig_repr__(self):
        return "required_callable"


required_callable = RequiredCallableClass()


class PythonFormatSpecificModel(MyModel):
    func: Callable[[], str]
    optional_func: Callable[[], str] = default_func


def get_pydantic_model_object() -> MyModel:
    all_kwargs = get_default_data()
    return MyModel(**all_kwargs)


def get_python_format_specific_model_object() -> PythonFormatSpecificModel:
    all_kwargs = get_default_data()
    all_kwargs["func"] = required_callable
    return PythonFormatSpecificModel(**all_kwargs)


@pytest.fixture
def pydantic_model_object() -> MyModel:
    yield get_pydantic_model_object()


@pytest.fixture
def python_format_specific_model_object() -> PythonFormatSpecificModel:
    yield get_python_format_specific_model_object()


PYTHON_FORMAT_IMPORT = "from tests.fixtures.pydantic_model import PythonFormatSpecificModel, SubModel, MyEnum, default_func, required_callable"



import functools
import platform
from enum import Enum


class OSType(str, Enum):
    WINDOWS = "Windows"
    LINUX = "Linux"
    MAC = "Mac"

    @property
    def name(self) -> str:
        if self.value == "Windows":
            return "Windows"
        elif self.value == "Linux":
            return "Linux"
        elif self.value == "Mac":
            return "Darwin"
        else:
            raise NotImplementedError(f"Unknown OS type: {self.value}")


def run_in(os_type: OSType):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if platform.system() != os_type.name:
                print(f"Skipping test {func.__name__} on non-{os_type.value} platform")
                return
            return func(*args, **kwargs)

        return wrapper

    return decorator



from enum import Enum
from pathlib import Path
from typing import Callable, Type

import pytest

from pyappconf import BaseConfig


class LoaderFunc(str, Enum):
    LOAD = "load"
    LOAD_OR_CREATE = "load_or_create"
    LOAD_RECURSIVE = "load_recursive"

    def get_loader(self, config_cls: Type[BaseConfig]) -> Callable:
        base_loader = getattr(config_cls, self.value)
        if self == LoaderFunc.LOAD_RECURSIVE:
            # Make recursive loader load the folder containing the config file
            def load_recursive_with_config_parent(path: Path, *args, **kwargs):
                return base_loader(path.parent, *args, **kwargs)

            return load_recursive_with_config_parent
        return base_loader


@pytest.fixture(params=list(LoaderFunc))
def loader_func(request) -> LoaderFunc:
    return request.param



import datetime
import re
import types
from enum import Enum
from pathlib import Path
from typing import Any, Sequence, Set
from uuid import UUID

import black
import isort
from pydantic import BaseModel
from pydantic.fields import ModelField


def pydantic_model_to_python_config_file(
    model: BaseModel,
    imports: Sequence[str],
    exclude_fields: Sequence[str] = tuple(),
) -> str:
    """
    Generate a python config file from a pydantic model.

    :param model: The pydantic model.
    :return: The python config file.
    """
    unformatted = _pydantic_model_to_python_config_file(model, imports, exclude_fields)
    # Format the python config file with black.
    formatted = _format_python_config_file(unformatted)
    return formatted


def pydantic_model_to_python_config_stub_file(
    model: BaseModel,
    imports: Sequence[str],
    exclude_fields: Sequence[str] = tuple(),
    generate_model_class: bool = False,
) -> str:
    """
    Generate the stub file for the generated python config file from a pydantic model.

    :param model: The pydantic model.
    :return: The python stub file.
    """
    unformatted = _pydantic_model_to_python_config_stub_file(
        model, imports, generate_model_class, exclude_fields
    )
    # Format the python config file with black.
    formatted = _format_python_config_file(unformatted, is_stub=True)
    return formatted


def _format_python_config_file(unformatted: str, is_stub: bool = False) -> str:
    """
    Format a python config file.

    :param unformatted: The python config file.
    :return: The formatted python config file.
    """
    isort_formatted = isort.code(unformatted)
    # Format with black second, so it can reformat the imports
    black_formatted = black.format_str(isort_formatted, mode=_black_file_mode(is_stub))

    return black_formatted


def _black_file_mode(is_stub: bool) -> black.FileMode:
    """
    Get the black file mode.

    :param is_stub: Whether the file is a stub.
    :return: The black file mode.
    """
    if is_stub:
        return black.FileMode(line_length=80, is_pyi=True)
    return black.FileMode(line_length=80)


def _pydantic_model_to_python_config_file(
    model: BaseModel,
    imports: Sequence[str],
    exclude_fields: Sequence[str] = tuple(),
) -> str:
    """
    Generate a python config file from a pydantic model.

    :param model: The pydantic model.
    :return: The python config file.
    """
    detected_stdlib_imports: Set[str] = set()
    # _build_attribute_value() will add imports to detected_stdlib_imports.
    attributes_str = _build_model_attributes(
        model, detected_stdlib_imports, exclude_fields
    )
    imports_str = "\n".join([*imports, *detected_stdlib_imports])
    name = model.__class__.__name__
    return f"""
{imports_str}

config = {name}({attributes_str})
""".strip()


def _pydantic_model_to_python_config_stub_file(
    model: BaseModel,
    imports: Sequence[str],
    generate_model_class: bool = False,
    exclude_fields: Sequence[str] = tuple(),
) -> str:
    """
    Generate a python config file from a pydantic model.

    :param model: The pydantic model.
    :return: The python config file.
    """
    name = model.__class__.__name__
    if not generate_model_class:
        imports_str = "\n".join(imports)
        return f"""
{imports_str}

config: {name}
""".strip()

    detected_stdlib_imports: Set[str] = set()
    # _build_attribute_value() and _build_attribute_type() will add imports to detected_stdlib_imports.
    attributes_str = _build_model_class_attributes(
        model, detected_stdlib_imports, exclude_fields
    )
    detected_stdlib_imports.add("from pydantic import BaseModel")
    imports_str = "\n".join([*imports, *detected_stdlib_imports])

    return f"""
{imports_str}

class {name}(BaseModel):
{attributes_str}

config: {name}
""".strip()


def _build_model_attributes(
    model: BaseModel, stdlib_imports: Set[str], exclude_fields: Sequence[str] = tuple()
) -> str:
    """
    Build the attribute definition of a pydantic model.

    :param model: The pydantic model.
    :return: The attributes of the model.
    """
    attributes = ""
    field: ModelField
    for field_name, field in model.__fields__.items():
        if field_name in exclude_fields:
            continue
        value = getattr(model, field_name)
        attributes += (
            f"    {field_name} = {_build_attribute_value(value, stdlib_imports)},\n"
        )
    return attributes


def _build_model_class_attributes(
    model: BaseModel, stdlib_imports: Set[str], exclude_fields: Sequence[str] = tuple()
) -> str:
    """
    Build the attribute definition of a pydantic model class.

    :param model: The pydantic model.
    :return: The attributes of the model class.
    """
    attributes = ""
    field: ModelField
    for field_name, field in model.__fields__.items():
        if field_name in exclude_fields:
            continue
        value = getattr(model, field_name)
        attributes += f"    {field_name}: {_build_attribute_type(field.type_, stdlib_imports)} = {_build_attribute_value(value, stdlib_imports)}\n"
    return attributes


TYPE_STR_RE = re.compile(r"<(class|enum) '([\w.]+)'>")


def _build_attribute_type(value: type, stdlib_imports: Set[str]) -> str:
    """
    Build the type of an attribute for a pydantic model class definition.

    :param value: A type class.
    :return: The code string representation of the type class.
    """
    str_value = str(value)
    # Parse the format <class 'str'> to str using a regular expression.
    match = TYPE_STR_RE.match(str_value)
    if match:
        fully_qualified_path = match.group(2)
        # fully_qualified_path could be just e.g. str or could be e.g. my.package.Something
        # We are assuming that the user is using a direct import for any custom types, so
        # extract only the name from the fully qualified path.
        import_name = fully_qualified_path.split(".")[-1]
        return import_name
    # May be in format typing.Callable[[], str]
    if "typing" in str_value:
        stdlib_imports.add("import typing")
        return str_value
    raise ValueError("Unsupported type class: " + str(value))


def _build_attribute_value(value: Any, stdlib_imports: Set[str]) -> str:
    """
    Build the attribute value of a pydantic model.

    :param value: The value of the attribute.
    :return: The value of the attribute.
    """
    if hasattr(value, "__pyconfig_repr__"):
        return value.__pyconfig_repr__()
    if isinstance(value, Enum):
        return f"{value.__class__.__name__}.{value.name}"
    elif isinstance(value, Path):
        stdlib_imports.add("from pathlib import Path")
        return f'Path("{value}")'
    elif isinstance(value, UUID):
        stdlib_imports.add("from uuid import UUID")
        return f'UUID("{value}")'
    elif isinstance(value, str):
        return f'"{value}"'
    elif isinstance(value, BaseModel):
        return repr(value)
    elif isinstance(value, (datetime.datetime, datetime.date, datetime.time)):
        stdlib_imports.add("import datetime")
        return repr(value)
    elif isinstance(value, types.FunctionType):
        return f"{value.__name__}"
    elif callable(value):
        raise ValueError(
            "Cannot convert generic callable to python config. Real functions defined with def are supported"
        )
    elif isinstance(value, list):
        return (
            f"[{', '.join(_build_attribute_value(v, stdlib_imports) for v in value)}]"
        )
    elif isinstance(value, tuple):
        return (
            f"({', '.join(_build_attribute_value(v, stdlib_imports) for v in value)})"
        )
    elif isinstance(value, dict):
        return f"{{{', '.join(f'{_build_attribute_value(k, stdlib_imports)}: {_build_attribute_value(v, stdlib_imports)}' for k, v in value.items())}}}"
    else:
        return str(value)



from pyappconf.py_config.generate import (
    pydantic_model_to_python_config_file,
    pydantic_model_to_python_config_stub_file,
)
from tests.config import PYDANTIC_PY_CONFIG_PATH, PYDANTIC_PY_CONFIG_STUB_PATH
from tests.fixtures.pydantic_model import (
    PYTHON_FORMAT_IMPORT,
    MyModel,
    python_format_specific_model_object,
)


def test_pydantic_model_to_config_file(python_format_specific_model_object: MyModel):
    config_str = pydantic_model_to_python_config_file(
        python_format_specific_model_object,
        [PYTHON_FORMAT_IMPORT],
    )
    assert config_str == PYDANTIC_PY_CONFIG_PATH.read_text()


def test_pydantic_model_to_stub_file(python_format_specific_model_object: MyModel):
    config_str = pydantic_model_to_python_config_stub_file(
        python_format_specific_model_object,
        [PYTHON_FORMAT_IMPORT],
        generate_model_class=True,
    )
    assert config_str == PYDANTIC_PY_CONFIG_STUB_PATH.read_text()



import os
from pathlib import Path

import pytest

from pyappconf import ConfigFormats
from tests.config import (
    JSON_PATH,
    PY_CONFIG_PATH,
    PYPROJECT_TOML_MIXED_CONFIG_PATH,
    TOML_PATH,
    YAML_PATH,
)
from tests.fixtures.loader_func import LoaderFunc, loader_func
from tests.fixtures.model import get_model_classes, get_model_object


@pytest.mark.parametrize("config_format", list(ConfigFormats))
def test_load_with_kwarg_and_env_overrides(
    config_format: ConfigFormats, loader_func: LoaderFunc
):
    config_path = _get_config_path_for_format(config_format)
    config_name = _get_config_name_for_format(config_format, config_path)
    os.environ["MYAPP_STRING"] = "abc"
    os.environ["MYAPP_INT_TUPLE"] = "[4, 5, 6]"

    OrigConfig, _ = get_model_classes()
    settings = OrigConfig._settings_with_overrides(
        default_format=config_format,
        custom_config_folder=config_path.parent,
        config_name=config_name,
    )

    class MyConfig(OrigConfig):
        _settings = settings

    model_object = get_model_object(
        string="abc", int_tuple=(40, 50, 60), integer=10, settings=settings
    )

    loader = loader_func.get_loader(MyConfig)
    loaded_object = loader(
        config_path, model_kwargs=dict(int_tuple=(40, 50, 60), integer=10)
    )

    del os.environ["MYAPP_STRING"]
    del os.environ["MYAPP_INT_TUPLE"]
    assert model_object == loaded_object

    # User passed should override env or config file
    # env should override config file
    # config file should be used if not overrided
    assert loaded_object.string == "abc"  # loaded from env, overriding config
    assert loaded_object.int_tuple == (
        40,
        50,
        60,
    )  # loaded from user kwarg, overriding env and config
    assert loaded_object.integer == 10  # loaded from user kwarg, overriding config
    assert loaded_object.str_list == ["a", "b", "c"]  # loaded from config


def _get_config_path_for_format(config_format: ConfigFormats) -> Path:
    if config_format == ConfigFormats.JSON:
        return JSON_PATH
    elif config_format == ConfigFormats.YAML:
        return YAML_PATH
    elif config_format == ConfigFormats.TOML:
        return TOML_PATH
    elif config_format == ConfigFormats.PY:
        return PY_CONFIG_PATH
    elif config_format == ConfigFormats.PYPROJECT:
        return PYPROJECT_TOML_MIXED_CONFIG_PATH
    else:
        raise NotImplementedError(f"{config_format} has no corresponding path")


def _get_config_name_for_format(config_format: ConfigFormats, config_path: Path) -> str:
    if config_format == ConfigFormats.PYPROJECT:
        return "config"
    else:
        return config_path.with_suffix("").name



from pathlib import Path

TEST_DIR = Path(__file__).parent
INPUT_DATA_DIR = TEST_DIR / "input_data"
GENERATED_DATA_DIR = TEST_DIR / "generated_data"

DATA_NAME = "data"
JSON_PATH = INPUT_DATA_DIR / (DATA_NAME + ".json")
JSON_WITH_SCHEMA_PATH = INPUT_DATA_DIR / (DATA_NAME + "-with-schema" + ".json")
YAML_PATH = INPUT_DATA_DIR / (DATA_NAME + ".yaml")
YAML_WITH_SCHEMA_PATH = INPUT_DATA_DIR / (DATA_NAME + "-with-schema" + ".yaml")
TOML_PATH = INPUT_DATA_DIR / (DATA_NAME + ".toml")
PY_CONFIG_PATH = INPUT_DATA_DIR / (DATA_NAME + ".py")
PY_CONFIG_STUB_PATH = PY_CONFIG_PATH.with_suffix(".pyi")

PYPROJECT_TOML_DIR = INPUT_DATA_DIR / "pyproject"
PYPROJECT_TOML_CLEAN_DIR = PYPROJECT_TOML_DIR / "clean"
PYPROJECT_TOML_CLEAN_CONFIG_PATH = PYPROJECT_TOML_CLEAN_DIR / "pyproject.toml"
PYPROJECT_TOML_MIXED_DIR = PYPROJECT_TOML_DIR / "mixed"
PYPROJECT_TOML_MIXED_CONFIG_PATH = PYPROJECT_TOML_MIXED_DIR / "pyproject.toml"

NON_EXISTENT_NAME = "woo"
NON_EXISTENT_IN_PATH = INPUT_DATA_DIR / NON_EXISTENT_NAME
NON_EXISTENT_INPUT_JSON_PATH = INPUT_DATA_DIR / (NON_EXISTENT_NAME + ".json")
CUSTOM_CONFIG_IN_PATH = INPUT_DATA_DIR / DATA_NAME
ENV_PATH = GENERATED_DATA_DIR / ".env"

RECURSIVE_INPUT_FOLDER = INPUT_DATA_DIR / "recursive"
RECURSIVE_CONFIG_1_PATH = RECURSIVE_INPUT_FOLDER / "1" / "recursive-config.toml"
RECURSIVE_CONFIG_ROOT_PATH = RECURSIVE_INPUT_FOLDER / "recursive-config.toml"

SCHEMA_JSON_PATH = INPUT_DATA_DIR / "schema.json"

PYDANTIC_PY_CONFIG_PATH = INPUT_DATA_DIR / "pydantic_py_config.py"
PYDANTIC_PY_CONFIG_STUB_PATH = PYDANTIC_PY_CONFIG_PATH.with_suffix(".pyi")

if not INPUT_DATA_DIR.exists():
    INPUT_DATA_DIR.mkdir()
if not GENERATED_DATA_DIR.exists():
    GENERATED_DATA_DIR.mkdir()



from enum import Enum
from typing import Optional, Sequence, Tuple, Type

import pytest
from pydantic import BaseModel

from pyappconf.model import AppConfig, BaseConfig, ConfigFormats
from tests.config import ENV_PATH, GENERATED_DATA_DIR
from tests.fixtures.data import get_default_data
from tests.fixtures.pydantic_model import MyModel, SubModel


class MyEnum(str, Enum):
    ONE = "one"
    TWO = "two"


class MyConfig(MyModel, BaseConfig):
    _settings: AppConfig = AppConfig(
        app_name="MyApp",
        py_config_imports=[
            "from tests.fixtures.model import MyConfig, SubModel, MyEnum"
        ],
    )

    class Config:
        env_prefix = "MYAPP_"
        env_file = ENV_PATH


class MyConfigPyFormat(MyConfig):
    _settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.PY,
        py_config_imports=[
            "from tests.fixtures.model import MyConfigPyFormat, SubModel, MyEnum"
        ],
    )


def get_model_classes() -> Tuple[Type[BaseConfig], Type[BaseModel]]:
    return MyConfig, SubModel


def get_model_class_with_defaults() -> Type[BaseConfig]:
    class MyDefaultConfig(BaseConfig):
        a: str = "a"
        b: int = 2

        _settings: AppConfig = AppConfig(app_name="MyApp")

        class Config:
            env_prefix = "MYAPP_"
            env_file = ENV_PATH

    return MyDefaultConfig


def get_model_object(
    exclude_keys: Optional[Sequence[str]] = None, **kwargs
) -> BaseConfig:
    MyConfig, SubModel = get_model_classes()
    all_kwargs = get_default_data(exclude_keys=exclude_keys, **kwargs)
    conf = MyConfig(**all_kwargs)
    return conf


def get_model_object_with_defaults() -> BaseConfig:
    MyDefaultConfig = get_model_class_with_defaults()
    return MyDefaultConfig()


@pytest.fixture(scope="session")
def model_classes() -> Tuple[Type[BaseConfig], Type[BaseModel]]:
    return get_model_classes()


@pytest.fixture(scope="session")
def model_class_with_defaults() -> Type[BaseConfig]:
    return get_model_class_with_defaults()


@pytest.fixture(scope="session")
def model_object() -> BaseConfig:
    return get_model_object()


@pytest.fixture(scope="session")
def model_object_with_defaults() -> BaseConfig:
    return get_model_object_with_defaults()



import shutil
from pathlib import Path
from typing import Optional, Tuple, Type

import pytest
from pydantic import BaseModel, create_model

from pyappconf.model import AppConfig, BaseConfig, ConfigFormats
from tests.config import (
    DATA_NAME,
    GENERATED_DATA_DIR,
    INPUT_DATA_DIR,
    JSON_PATH,
    NON_EXISTENT_INPUT_JSON_PATH,
    NON_EXISTENT_NAME,
    PYPROJECT_TOML_MIXED_CONFIG_PATH,
    RECURSIVE_INPUT_FOLDER,
    TOML_PATH,
    YAML_PATH,
)
from tests.fixtures.data import get_default_data
from tests.fixtures.model import (
    MyConfig,
    MyConfigPyFormat,
    SubModel,
    get_model_classes,
    get_model_object,
    model_class_with_defaults,
    model_classes,
    model_object,
    model_object_with_defaults,
)
from tests.fixtures.temp_folder import temp_folder


def _save_load_test(custom_settings: AppConfig) -> Tuple[BaseConfig, Type[BaseConfig]]:
    mod = get_model_object(settings=custom_settings)
    mod.save()
    assert str(mod.settings.config_location).endswith(
        custom_settings.default_format.value
    )

    OrigConfig, SubModel = get_model_classes()

    class MyConfig(OrigConfig):
        _settings = custom_settings

    assert str(MyConfig._settings.config_location).endswith(
        custom_settings.default_format.value
    )
    obj = MyConfig.load()
    # Check that loaded config is the same as the one used for saving
    assert mod == obj
    return obj, MyConfig


def test_save_load_toml():
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.TOML,
    )
    _save_load_test(custom_settings)


def test_save_load_yaml():
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.YAML,
    )
    _save_load_test(custom_settings)


def test_save_load_json():
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.JSON,
    )
    _save_load_test(custom_settings)


def test_save_load_py_config():
    """
    This is structured differently than other save/load tests because py config functionality
    will not work correctly when dynamically modifying the class settings just before load.
    py config uses the imported class to create the config object, and that imported class
    does not have the dynamic modifications applied.
    :return:
    """
    all_kwargs = get_default_data()
    mod = MyConfigPyFormat(**all_kwargs)
    mod.save()
    settings = MyConfigPyFormat._settings
    assert str(mod.settings.config_location).endswith(settings.default_format.value)

    obj = MyConfigPyFormat.load()
    # Check that loaded config is the same as the one used for saving
    assert mod == obj


def test_save_load_pyproject_toml(temp_folder: Path):
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        default_format=ConfigFormats.PYPROJECT,
    )
    _save_load_test(custom_settings)


def test_saving_pyproject_toml_does_not_affect_other_pyproject_content(
    model_object: MyConfig,
):
    path = PYPROJECT_TOML_MIXED_CONFIG_PATH
    orig_contents = path.read_text()
    last_modified = path.stat().st_mtime
    mod = MyConfig.load(path)
    assert mod == model_object.copy(
        update=dict(
            settings=model_object.settings.copy(
                default_format=ConfigFormats.PYPROJECT, custom_config_folder=path.parent
            )
        )
    )
    mod.save()
    assert path.read_text() == orig_contents
    # Assert config file was written to
    assert path.stat().st_mtime != last_modified


def test_save_load_yaml_with_schema():
    expect_schema_url = "https://example.com/schema.json"
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.YAML,
        schema_url=expect_schema_url,
    )
    obj, MyConfig = _save_load_test(custom_settings)
    assert MyConfig._settings.schema_url == expect_schema_url
    yaml_str = obj.to_yaml()
    assert expect_schema_url in yaml_str
    yaml_str_no_schema = obj.to_yaml(include_schema_url=False)
    assert expect_schema_url not in yaml_str_no_schema


def test_save_load_json_with_schema():
    expect_schema_url = "https://example.com/schema.json"
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        default_format=ConfigFormats.JSON,
        schema_url=expect_schema_url,
    )
    obj, MyConfig = _save_load_test(custom_settings)
    assert MyConfig._settings.schema_url == expect_schema_url
    json_str = obj.to_json()
    assert expect_schema_url in json_str
    json_str_no_schema = obj.to_json(include_schema_url=False)
    assert expect_schema_url not in json_str_no_schema


def _multi_format_save_load_test(
    custom_settings: AppConfig,
) -> Tuple[BaseConfig, Type[BaseConfig]]:
    # Save in a format other than the default format
    settings_with_other_format: Optional[AppConfig] = None
    for config_format in ConfigFormats:
        if config_format == custom_settings.default_format:
            continue
        settings_with_other_format = custom_settings.copy(default_format=config_format)
    if settings_with_other_format is None:
        raise ValueError("No other formats to test")

    mod = get_model_object(settings=custom_settings)
    mod.save()
    assert str(mod.settings.config_location).endswith(
        custom_settings.default_format.value
    )

    OrigConfig, SubModel = get_model_classes()

    # Now MyConfig is configured with a different format, so it could only be loaded
    # via multi-format
    class MyConfig(OrigConfig):
        _settings = settings_with_other_format

    assert str(MyConfig._settings.config_location).endswith(
        settings_with_other_format.default_format.value
    )
    obj = MyConfig.load(multi_format=True)
    # Check that loaded config is the same as the one used for saving
    assert obj == mod.copy(
        update=dict(
            settings=mod.settings.copy(default_format=custom_settings.default_format)
        )
    )
    return obj, MyConfig


def test_multi_format_save_load_toml(temp_folder: Path):
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        default_format=ConfigFormats.TOML,
        py_config_imports=[
            "from tests.fixtures.model import MyConfig, SubModel, MyEnum"
        ],
    )
    _multi_format_save_load_test(custom_settings)


def test_multi_format_save_load_yaml(temp_folder: Path):
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        default_format=ConfigFormats.YAML,
        py_config_imports=[
            "from tests.fixtures.model import MyConfig, SubModel, MyEnum"
        ],
    )
    _multi_format_save_load_test(custom_settings)


def test_multi_format_save_load_json(temp_folder: Path):
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        default_format=ConfigFormats.JSON,
        py_config_imports=[
            "from tests.fixtures.model import MyConfig, SubModel, MyEnum"
        ],
    )
    _multi_format_save_load_test(custom_settings)


def test_multi_format_save_load_py_config(temp_folder: Path):
    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        default_format=ConfigFormats.PY,
        config_name="with-dashes and spaces",
        py_config_imports=[
            "from tests.fixtures.model import MyConfig, SubModel, MyEnum"
        ],
    )
    _multi_format_save_load_test(custom_settings)


def assert_model_loaded_with_extension(
    mod: BaseConfig, model_object: BaseConfig, config_format: ConfigFormats
):
    assert mod.settings.custom_config_folder == INPUT_DATA_DIR
    assert mod.settings.default_format == config_format
    assert mod.settings.config_name == DATA_NAME

    mod.settings = model_object.settings
    assert mod == model_object


def test_load_toml_with_custom_path(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load(TOML_PATH)
    assert mod._settings.config_location == orig_config_location
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.TOML)


def test_load_yaml_with_custom_path(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load(YAML_PATH)
    assert mod._settings.config_location == orig_config_location
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.YAML)


def test_load_json_with_custom_path(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load(JSON_PATH)
    assert mod._settings.config_location == orig_config_location
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.JSON)


def test_load_or_create_with_path_exists(
    model_object: BaseConfig, model_classes: Tuple[Type[BaseConfig], Type[BaseModel]]
):
    MyConfig, _ = model_classes
    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load_or_create(JSON_PATH)
    assert mod._settings.config_location == orig_config_location
    assert_model_loaded_with_extension(mod, model_object, ConfigFormats.JSON)


def test_load_or_create_with_path_does_not_exist(
    model_object_with_defaults: BaseConfig, model_class_with_defaults: Type[BaseConfig]
):
    MyConfig = model_class_with_defaults
    model_object = model_object_with_defaults
    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load_or_create(NON_EXISTENT_INPUT_JSON_PATH)
    assert mod._settings.config_location == orig_config_location
    assert mod.settings.custom_config_folder == INPUT_DATA_DIR
    assert mod.settings.default_format == ConfigFormats.JSON
    assert mod.settings.config_name == NON_EXISTENT_NAME

    mod.settings = model_object.settings
    assert mod == model_object


def test_load_or_create_with_folder_and_no_multi_format(
    model_object_with_defaults: BaseConfig,
    model_class_with_defaults: Type[BaseConfig],
    temp_folder: Path,
):
    MyConfig = model_class_with_defaults
    model_object = model_object_with_defaults
    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load_or_create(temp_folder)
    assert mod._settings.config_location == orig_config_location
    assert mod.settings.custom_config_folder == temp_folder
    assert mod.settings.default_format == ConfigFormats.TOML
    assert mod.settings.config_name == "config"

    mod.settings = model_object.settings
    assert mod == model_object


def test_load_or_create_with_folder_and_multi_format_and_file_exists(
    model_object: BaseConfig,
    model_classes: Tuple[Type[BaseConfig], Type[BaseModel]],
    temp_folder: Path,
):
    MyConfig, _ = model_classes
    orig_config_location = MyConfig._settings.config_location
    shutil.copy(JSON_PATH, temp_folder / "config.json")
    mod = MyConfig.load_or_create(temp_folder, multi_format=True)
    assert mod._settings.config_location == orig_config_location
    assert mod.settings.custom_config_folder == temp_folder
    assert mod.settings.default_format == ConfigFormats.JSON
    assert mod.settings.config_name == "config"
    assert mod == model_object.copy(
        update=dict(
            settings=model_object.settings.copy(
                custom_config_folder=temp_folder, default_format=ConfigFormats.JSON
            )
        )
    )


def test_load_or_create_with_folder_and_multi_format_and_file_does_not_exist(
    model_object_with_defaults: BaseConfig,
    model_class_with_defaults: Type[BaseConfig],
    temp_folder: Path,
):
    MyConfig = model_class_with_defaults
    model_object = model_object_with_defaults

    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load_or_create(temp_folder, multi_format=True)
    assert mod._settings.config_location == orig_config_location
    assert mod.settings.custom_config_folder == temp_folder
    assert mod.settings.default_format == ConfigFormats.TOML
    assert mod.settings.config_name == "config"
    assert mod == model_object.copy(
        update=dict(
            settings=model_object.settings.copy(custom_config_folder=temp_folder)
        )
    )


@pytest.mark.parametrize(
    "path, expect_string",
    [
        (RECURSIVE_INPUT_FOLDER / "1", "loaded from 1"),
        (RECURSIVE_INPUT_FOLDER / "2", "loaded from recursive"),
        # Third and fourth cases - didn't find recursive, go to default location
        (INPUT_DATA_DIR, "a"),
        (None, "a"),
    ],
)
def test_load_recursive(path: Optional[Path], expect_string: str):
    OrigConfig, SubModel = get_model_classes()

    shutil.copy(
        INPUT_DATA_DIR / "data.toml", GENERATED_DATA_DIR / "recursive-config.toml"
    )

    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=GENERATED_DATA_DIR,
        config_name="recursive-config",
    )

    class MyConfig(OrigConfig):
        _settings = custom_settings

    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load_recursive(path, multi_format=False)
    assert mod._settings.config_location == orig_config_location

    assert mod.string == expect_string


@pytest.mark.parametrize(
    "config_format", [ConfigFormats.JSON, ConfigFormats.YAML, ConfigFormats.PY]
)
def test_load_recursive_multi_format(config_format: ConfigFormats, temp_folder: Path):
    OrigConfig, SubModel = get_model_classes()

    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        # Set to yaml so that toml could only be loaded via multi_format
        default_format=config_format,
        config_name="recursive-config",
    )

    class MyConfig(OrigConfig):
        _settings = custom_settings

    orig_config_location = MyConfig._settings.config_location
    mod = MyConfig.load_recursive(RECURSIVE_INPUT_FOLDER / "2", multi_format=True)
    assert mod._settings.config_location == orig_config_location

    assert mod.string == "loaded from recursive"


def test_save_load_dynamic(temp_folder: Path):
    OrigConfig, SubModel = get_model_classes()

    custom_settings = AppConfig(
        app_name="MyApp",
        custom_config_folder=temp_folder,
        config_name="dynamic-config",
    )

    class MyConfig(OrigConfig):
        _settings = custom_settings

    DynamicModel = create_model(
        "DynamicModel", dynamic_attr=(str, "dynamic"), integer=100, __base__=MyConfig
    )

    data = {**get_default_data(), **{"integer": 100}}
    model = DynamicModel(**data)
    model.save()

    loaded = DynamicModel.load()
    assert model == loaded
    assert model.dynamic_attr == "dynamic"
    assert model.integer == 100
    assert model.string == "a"



import shutil
from pathlib import Path

from pyappconf import ConfigFormats
from tests.config import (
    JSON_PATH,
    PY_CONFIG_PATH,
    PYPROJECT_TOML_CLEAN_CONFIG_PATH,
    TOML_PATH,
    YAML_PATH,
)
from tests.fixtures.model import MyConfig, model_object
from tests.fixtures.temp_folder import temp_folder


def test_load_yaml_via_multi_format(temp_folder: Path, model_object: MyConfig):
    shutil.copy(YAML_PATH, temp_folder / "config.yaml")

    config = MyConfig.load(temp_folder, multi_format=True)
    expect_altered_settings = MyConfig._settings_with_overrides(
        custom_config_folder=temp_folder, default_format=ConfigFormats.YAML
    )
    assert config == model_object.copy(update=dict(settings=expect_altered_settings))


def test_load_json_via_multi_format(temp_folder: Path, model_object: MyConfig):
    shutil.copy(JSON_PATH, temp_folder / "config.json")

    config = MyConfig.load(temp_folder, multi_format=True)
    expect_altered_settings = MyConfig._settings_with_overrides(
        custom_config_folder=temp_folder, default_format=ConfigFormats.JSON
    )
    assert config == model_object.copy(update=dict(settings=expect_altered_settings))


def test_load_toml_via_multi_format(temp_folder: Path, model_object: MyConfig):
    shutil.copy(TOML_PATH, temp_folder / "config.toml")

    config = MyConfig.load(temp_folder, multi_format=True)
    expect_altered_settings = MyConfig._settings_with_overrides(
        custom_config_folder=temp_folder, default_format=ConfigFormats.TOML
    )
    assert config == model_object.copy(update=dict(settings=expect_altered_settings))


def test_load_py_config_via_multi_format(temp_folder: Path, model_object: MyConfig):
    shutil.copy(PY_CONFIG_PATH, temp_folder / "config.py")

    config = MyConfig.load(temp_folder, multi_format=True)
    expect_altered_settings = MyConfig._settings_with_overrides(
        custom_config_folder=temp_folder, default_format=ConfigFormats.PY
    )
    assert config == model_object.copy(update=dict(settings=expect_altered_settings))


def test_load_pyproject_toml_via_multi_format(
    temp_folder: Path, model_object: MyConfig
):
    shutil.copy(PYPROJECT_TOML_CLEAN_CONFIG_PATH, temp_folder / "pyproject.toml")

    config = MyConfig.load(temp_folder, multi_format=True)
    expect_altered_settings = MyConfig._settings_with_overrides(
        custom_config_folder=temp_folder, default_format=ConfigFormats.PYPROJECT
    )
    assert config == model_object.copy(update=dict(settings=expect_altered_settings))



"""
Generate test data for file output
"""
from pyappconf import AppConfig
from pyappconf.py_config.generate import (
    pydantic_model_to_python_config_file,
    pydantic_model_to_python_config_stub_file,
)
from tests.config import (
    DATA_NAME,
    INPUT_DATA_DIR,
    JSON_PATH,
    JSON_WITH_SCHEMA_PATH,
    PY_CONFIG_PATH,
    PYDANTIC_PY_CONFIG_PATH,
    PYDANTIC_PY_CONFIG_STUB_PATH,
    PYPROJECT_TOML_CLEAN_CONFIG_PATH,
    PYPROJECT_TOML_MIXED_CONFIG_PATH,
    RECURSIVE_CONFIG_1_PATH,
    RECURSIVE_CONFIG_ROOT_PATH,
    SCHEMA_JSON_PATH,
    TOML_PATH,
    YAML_PATH,
    YAML_WITH_SCHEMA_PATH,
)
from tests.fixtures.model import MyConfig, get_model_object
from tests.fixtures.pydantic_model import (
    PYTHON_FORMAT_IMPORT,
    get_pydantic_model_object,
    get_python_format_specific_model_object,
)


def generate_basic_configs():
    conf = get_model_object()
    conf.to_json(out_path=JSON_PATH, json_kwargs=dict(indent=2))
    conf.to_yaml(out_path=YAML_PATH)
    conf.to_toml(out_path=TOML_PATH)
    conf.to_py_config(out_path=PY_CONFIG_PATH)
    conf.to_pyproject_toml(out_path=PYPROJECT_TOML_CLEAN_CONFIG_PATH)
    conf.to_pyproject_toml(out_path=PYPROJECT_TOML_MIXED_CONFIG_PATH)


def generate_configs_with_schema():
    settings_with_schema = AppConfig(
        app_name="MyApp", schema_url="https://example.com/schema.json"
    )
    conf_with_schema = get_model_object(settings=settings_with_schema)
    conf_with_schema.to_json(out_path=JSON_WITH_SCHEMA_PATH, json_kwargs=dict(indent=2))
    conf_with_schema.to_yaml(out_path=YAML_WITH_SCHEMA_PATH)


def generate_recursive_configs():
    config_1 = get_model_object(string="loaded from 1")
    config_1.to_toml(RECURSIVE_CONFIG_1_PATH)
    config_recursive = get_model_object(string="loaded from recursive")
    config_recursive.to_toml(RECURSIVE_CONFIG_ROOT_PATH)


def generate_json_schema():
    schema = MyConfig.schema_json(indent=2)
    SCHEMA_JSON_PATH.write_text(schema)


def generate_pydantic_py_config_file():
    model = get_python_format_specific_model_object()
    conf_str = pydantic_model_to_python_config_file(
        model,
        [PYTHON_FORMAT_IMPORT],
    )
    PYDANTIC_PY_CONFIG_PATH.write_text(conf_str)


def generate_pydantic_py_config_stub_file():
    model = get_python_format_specific_model_object()
    conf_str = pydantic_model_to_python_config_stub_file(
        model, [PYTHON_FORMAT_IMPORT], generate_model_class=True
    )
    PYDANTIC_PY_CONFIG_STUB_PATH.write_text(conf_str)


if __name__ == "__main__":
    generate_basic_configs()
    generate_configs_with_schema()
    generate_recursive_configs()
    generate_json_schema()
    generate_pydantic_py_config_file()
    generate_pydantic_py_config_stub_file()



from pathlib import Path

from pyappconf import AppConfig


def test_copy_settings():
    app_config = AppConfig(app_name="MyApp")
    orig_location = app_config.config_location
    new_folder = Path("somewhere")
    new_location = new_folder / "config.toml"
    new_config = app_config.copy(custom_config_folder=new_folder)
    assert new_config.config_location == new_location
    assert app_config.config_location == orig_location



import datetime
import json
from pathlib import Path
from typing import Any
from uuid import UUID


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, (Path, UUID)):
            return str(o)

        if isinstance(o, (datetime.date, datetime.datetime)):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)
